
sphinx_conf = r"""
import sys
sys.path.insert(0, '..')
sys.path.insert(0, '.')

project = "TITLE"
copyright = "COPYRIGHT"
author = "AUTHOR"
release = "RELEASE"

extensions = [EXTENSIONS]

templates_path = ['_templates']
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']
master_doc = 'INDEX'
numfig = True
numfig_secnum_depth = 2
math_numfig = True
math_number_all = True

suppress_warnings = ['misc.highlighting_failure']
linkcheck_ignore = [r'.*localhost.*']
linkcheck_timeout = 5
linkcheck_workers = 20

html_theme = 'mxtheme'
html_theme_options = {
    'primary_color': 'blue',
    'accent_color': 'deep_orange',
    'header_links': [
        HEADER_LINKS
    ],
    'show_footer': False
}
html_static_path = ['_static']

html_favicon = 'FAVICON'

html_logo = 'HTML_LOGO'

latex_documents = [
    (master_doc, "NAME.tex", "TITLE",
     author, 'manual'),
]

rsvg_converter_args = ['-z', '0.8']

latex_engine = 'xelatex' # for utf-8 supports
latex_show_pagerefs = True
latex_show_urls = 'footnote'

latex_logo = 'LATEX_LOGO'

latex_elements = {

'figure_align': 'H',

'pointsize': '11pt',
'preamble': r'''

% Page size
\setlength{\voffset}{-14mm}
\addtolength{\textheight}{16mm}

% Chapter title style
\usepackage{titlesec, blindtext, color}
\definecolor{gray75}{gray}{0.75}
\newcommand{\hsp}{\hspace{20pt}}
\titleformat{\chapter}[hang]{\Huge\bfseries}{\thechapter\hsp\textcolor{gray75}{|}\hsp}{0pt}{\Huge\bfseries}

% So some large pictures won't get the full page
\renewcommand{\floatpagefraction}{.8}

\setcounter{tocdepth}{1}
% Use natbib's citation style, e.g. (Li and Smola, 16)
\usepackage{natbib}
\protected\def\sphinxcite{\citep}

MAIN_FONT
SANS_FONT
MONO_FONT

% Remove top header
\usepackage[draft]{minted}
\fvset{breaklines=true, breakanywhere=true}
\setlength{\headheight}{13.6pt}
\makeatletter
    \fancypagestyle{normal}{
        \fancyhf{}
        \fancyfoot[LE,RO]{{\py@HeaderFamily\thepage}}
        \fancyfoot[LO]{{\py@HeaderFamily\nouppercase{\rightmark}}}
        \fancyfoot[RE]{{\py@HeaderFamily\nouppercase{\leftmark}}}
        \fancyhead[LE,RO]{{\py@HeaderFamily }}
     }
\makeatother
''',
'sphinxsetup': 'verbatimwithframe=false, verbatimsep=2mm, VerbatimColor={rgb}{.95,.95,.95}'
}

SPHINX_CONFIGS

def setup(app):
    # app.add_javascript('https://cdnjs.cloudflare.com/ajax/libs/clipboard.js/2.0.0/clipboard.min.js')
    app.add_javascript('d2l.js')
    app.add_stylesheet('d2l.css')
    import mxtheme
    app.add_directive('card', mxtheme.CardDirective)
"""


google_tracker = """
(function(i,s,o,g,r,a,m){i['GoogleAnalyticsObject']=r;i[r]=i[r]||function(){
(i[r].q=i[r].q||[]).push(arguments)},i[r].l=1*new Date();a=s.createElement(o),
m=s.getElementsByTagName(o)[0];a.async=1;a.src=g;m.parentNode.insertBefore(a,m)
})(window,document,'script','https://www.google-analytics.com/analytics.js','ga');

ga('create', 'GOOGLE_ANALYTICS_TRACKING_ID', 'auto');
ga('send', 'pageview');
var captureOutboundLink = function(url) {
   ga('send', 'event', 'outbound', 'click', url, {
     'transport': 'beacon',
     'hitCallback': function(){document.location = url;}
   });
}

var tagClick = function(tab) {
    ga('send', 'event', 'tab', 'click', tab, {
        'transport': 'beacon'
    });
}
"""

shorten_sec_num = """
$(document).ready(function () {
    $('.localtoc').each(function(){
        $(this).find('a').each(function(){
            $(this).html($(this).html().replace(/^\d+\.\d+\./, ''))
        });
    });
});
"""

# Replace the QR code with an embeded discussion thread.
replace_qr = """
$(document).ready(function () {
    $('h2').each(function(){
        if ($(this).text().indexOf("Discuss") != -1) {
            var url = $(this).find('a').attr('href');
            var tokens = url.split('/');
            var topic_id = tokens[tokens.length-1];
            var domain = tokens[0]+'//'+tokens[2]+'/';
            $(this).parent().append('<div id="discourse-comments"></div>');

            $('img').each(function(){
                if ($(this).attr('src').indexOf("qr_") != -1) {
                    $(this).hide();
                }
            });

            DiscourseEmbed = { discourseUrl: domain, topicId: topic_id };
            (function() {
                var d = document.createElement('script'); d.type = 'text/javascript';
                d.async = true;
                d.src = DiscourseEmbed.discourseUrl + 'javascripts/embed.js';
                (document.getElementsByTagName('head')[0] ||
                 document.getElementsByTagName('body')[0]).appendChild(d);
            })();
        }
    });
});
"""

# CSS style to hide the bibkey but allow highlight
hide_bibkey_css = r"""
dl.citation > dt.label {
    width: 100%;
}

dl.citation span.brackets {
    padding-right: .8em;
}
"""

discourse_js = r"""
function discourse_embed() {
    $('a').each(function(){
        if ($(this).text().indexOf("Discussions") != -1) {
            var pp = $(this).parent();
            if (pp.is('h2')  || pp.is('li')) {
                return;
            }
            var pp = $(this).parent().parent();
            if (pp.hasClass('mdl-tabs__panel') && !pp.hasClass('is-active')) {
                return;
            }
            var url = $(this).attr('href');
            var tokens = url.split('/');
            var topic_id = tokens[tokens.length-1];
            var domain = tokens[0]+'//'+tokens[2]+'/';
            $("#discourse-comments").remove();
            $(this).parent().append('<div id="discourse-comments"></div>');


            DiscourseEmbed = { discourseUrl: domain, topicId: topic_id };
            (function() {
                var d = document.createElement('script'); d.type = 'text/javascript';
                d.async = true;
                d.src = DiscourseEmbed.discourseUrl + 'javascripts/embed.js';
                (document.getElementsByTagName('head')[0] ||
                 document.getElementsByTagName('body')[0]).appendChild(d);
            })();
            $(this).hide()
        }
    });
}
"""

tabbar_js = r"""
function select_tab(i) {
    var n = 4;
    $(".mdl-tabs").each(function(index){
        var j;
        for (j = 0; j < n; j++) {
            if (j != i) {
                $(this).find(".mdl-tabs__panel:eq("+j.toString()+")").removeClass('is-active');
            }
        }
        $(this).find(".mdl-tabs__panel:eq("+i.toString()+")").addClass('is-active');
    });
    $(".d2l-tabs").each(function(index){
        var j;
        for (j = 0; j < n; j++) {
            if (j != i) {
                $(this).find(".d2l-tabs__tab:eq("+j.toString()+")").hide();
            }
        }
        $(this).find(".d2l-tabs__tab:eq("+i.toString()+")").show();
    });
    $(".mdl-tabs__tab-bar").each(function(index){
        var j;
        for (j = 0; j < n; j++) {
            if (j != i) {
                $(this).find(".mdl-tabs__tab:eq("+j.toString()+")").removeClass('is-active');
            }
        }
        $(this).find(".mdl-tabs__tab:eq("+i.toString()+")").addClass('is-active');
    });
    discourse_embed();
}

$(document).ready(function () {
  select_tab(0);
  $(".mdl-tabs__tab-bar").each(function(index){
    $(this).find(".mdl-tabs__tab:eq(0)").click(function() {
        select_tab(0);
    });
    $(this).find(".mdl-tabs__tab:eq(1)").click(function() {
        select_tab(1);
    });
    $(this).find(".mdl-tabs__tab:eq(2)").click(function() {
        select_tab(2);
    });
    $(this).find(".mdl-tabs__tab:eq(3)").click(function() {
        select_tab(3);
    });
  });
});
"""
copybutton_js = r"""

!function(t,e){"object"==typeof exports&&"object"==typeof module?module.exports=e():"function"==typeof define&&define.amd?define([],e):"object"==typeof exports?exports.ClipboardJS=e():t.ClipboardJS=e()}(this,function(){return function(t){function e(o){if(n[o])return n[o].exports;var r=n[o]={i:o,l:!1,exports:{}};return t[o].call(r.exports,r,r.exports,e),r.l=!0,r.exports}var n={};return e.m=t,e.c=n,e.i=function(t){return t},e.d=function(t,n,o){e.o(t,n)||Object.defineProperty(t,n,{configurable:!1,enumerable:!0,get:o})},e.n=function(t){var n=t&&t.__esModule?function(){return t.default}:function(){return t};return e.d(n,"a",n),n},e.o=function(t,e){return Object.prototype.hasOwnProperty.call(t,e)},e.p="",e(e.s=3)}([function(t,e,n){var o,r,i;!function(a,c){r=[t,n(7)],o=c,void 0!==(i="function"==typeof o?o.apply(e,r):o)&&(t.exports=i)}(0,function(t,e){"use strict";function n(t,e){if(!(t instanceof e))throw new TypeError("Cannot call a class as a function")}var o=function(t){return t&&t.__esModule?t:{default:t}}(e),r="function"==typeof Symbol&&"symbol"==typeof Symbol.iterator?function(t){return typeof t}:function(t){return t&&"function"==typeof Symbol&&t.constructor===Symbol&&t!==Symbol.prototype?"symbol":typeof t},i=function(){function t(t,e){for(var n=0;n<e.length;n++){var o=e[n];o.enumerable=o.enumerable||!1,o.configurable=!0,"value"in o&&(o.writable=!0),Object.defineProperty(t,o.key,o)}}return function(e,n,o){return n&&t(e.prototype,n),o&&t(e,o),e}}(),a=function(){function t(e){n(this,t),this.resolveOptions(e),this.initSelection()}return i(t,[{key:"resolveOptions",value:function(){var t=arguments.length>0&&void 0!==arguments[0]?arguments[0]:{};this.action=t.action,this.container=t.container,this.emitter=t.emitter,this.target=t.target,this.text=t.text,this.trigger=t.trigger,this.selectedText=""}},{key:"initSelection",value:function(){this.text?this.selectFake():this.target&&this.selectTarget()}},{key:"selectFake",value:function(){var t=this,e="rtl"==document.documentElement.getAttribute("dir");this.removeFake(),this.fakeHandlerCallback=function(){return t.removeFake()},this.fakeHandler=this.container.addEventListener("click",this.fakeHandlerCallback)||!0,this.fakeElem=document.createElement("textarea"),this.fakeElem.style.fontSize="12pt",this.fakeElem.style.border="0",this.fakeElem.style.padding="0",this.fakeElem.style.margin="0",this.fakeElem.style.position="absolute",this.fakeElem.style[e?"right":"left"]="-9999px";var n=window.pageYOffset||document.documentElement.scrollTop;this.fakeElem.style.top=n+"px",this.fakeElem.setAttribute("readonly",""),this.fakeElem.value=this.text,this.container.appendChild(this.fakeElem),this.selectedText=(0,o.default)(this.fakeElem),this.copyText()}},{key:"removeFake",value:function(){this.fakeHandler&&(this.container.removeEventListener("click",this.fakeHandlerCallback),this.fakeHandler=null,this.fakeHandlerCallback=null),this.fakeElem&&(this.container.removeChild(this.fakeElem),this.fakeElem=null)}},{key:"selectTarget",value:function(){this.selectedText=(0,o.default)(this.target),this.copyText()}},{key:"copyText",value:function(){var t=void 0;try{t=document.execCommand(this.action)}catch(e){t=!1}this.handleResult(t)}},{key:"handleResult",value:function(t){this.emitter.emit(t?"success":"error",{action:this.action,text:this.selectedText,trigger:this.trigger,clearSelection:this.clearSelection.bind(this)})}},{key:"clearSelection",value:function(){this.trigger&&this.trigger.focus(),window.getSelection().removeAllRanges()}},{key:"destroy",value:function(){this.removeFake()}},{key:"action",set:function(){var t=arguments.length>0&&void 0!==arguments[0]?arguments[0]:"copy";if(this._action=t,"copy"!==this._action&&"cut"!==this._action)throw new Error('Invalid "action" value, use either "copy" or "cut"')},get:function(){return this._action}},{key:"target",set:function(t){if(void 0!==t){if(!t||"object"!==(void 0===t?"undefined":r(t))||1!==t.nodeType)throw new Error('Invalid "target" value, use a valid Element');if("copy"===this.action&&t.hasAttribute("disabled"))throw new Error('Invalid "target" attribute. Please use "readonly" instead of "disabled" attribute');if("cut"===this.action&&(t.hasAttribute("readonly")||t.hasAttribute("disabled")))throw new Error('Invalid "target" attribute. You can\'t cut text from elements with "readonly" or "disabled" attributes');this._target=t}},get:function(){return this._target}}]),t}();t.exports=a})},function(t,e,n){function o(t,e,n){if(!t&&!e&&!n)throw new Error("Missing required arguments");if(!c.string(e))throw new TypeError("Second argument must be a String");if(!c.fn(n))throw new TypeError("Third argument must be a Function");if(c.node(t))return r(t,e,n);if(c.nodeList(t))return i(t,e,n);if(c.string(t))return a(t,e,n);throw new TypeError("First argument must be a String, HTMLElement, HTMLCollection, or NodeList")}function r(t,e,n){return t.addEventListener(e,n),{destroy:function(){t.removeEventListener(e,n)}}}function i(t,e,n){return Array.prototype.forEach.call(t,function(t){t.addEventListener(e,n)}),{destroy:function(){Array.prototype.forEach.call(t,function(t){t.removeEventListener(e,n)})}}}function a(t,e,n){return u(document.body,t,e,n)}var c=n(6),u=n(5);t.exports=o},function(t,e){function n(){}n.prototype={on:function(t,e,n){var o=this.e||(this.e={});return(o[t]||(o[t]=[])).push({fn:e,ctx:n}),this},once:function(t,e,n){function o(){r.off(t,o),e.apply(n,arguments)}var r=this;return o._=e,this.on(t,o,n)},emit:function(t){var e=[].slice.call(arguments,1),n=((this.e||(this.e={}))[t]||[]).slice(),o=0,r=n.length;for(o;o<r;o++)n[o].fn.apply(n[o].ctx,e);return this},off:function(t,e){var n=this.e||(this.e={}),o=n[t],r=[];if(o&&e)for(var i=0,a=o.length;i<a;i++)o[i].fn!==e&&o[i].fn._!==e&&r.push(o[i]);return r.length?n[t]=r:delete n[t],this}},t.exports=n},function(t,e,n){var o,r,i;!function(a,c){r=[t,n(0),n(2),n(1)],o=c,void 0!==(i="function"==typeof o?o.apply(e,r):o)&&(t.exports=i)}(0,function(t,e,n,o){"use strict";function r(t){return t&&t.__esModule?t:{default:t}}function i(t,e){if(!(t instanceof e))throw new TypeError("Cannot call a class as a function")}function a(t,e){if(!t)throw new ReferenceError("this hasn't been initialised - super() hasn't been called");return!e||"object"!=typeof e&&"function"!=typeof e?t:e}function c(t,e){if("function"!=typeof e&&null!==e)throw new TypeError("Super expression must either be null or a function, not "+typeof e);t.prototype=Object.create(e&&e.prototype,{constructor:{value:t,enumerable:!1,writable:!0,configurable:!0}}),e&&(Object.setPrototypeOf?Object.setPrototypeOf(t,e):t.__proto__=e)}function u(t,e){var n="data-clipboard-"+t;if(e.hasAttribute(n))return e.getAttribute(n)}var l=r(e),s=r(n),f=r(o),d="function"==typeof Symbol&&"symbol"==typeof Symbol.iterator?function(t){return typeof t}:function(t){return t&&"function"==typeof Symbol&&t.constructor===Symbol&&t!==Symbol.prototype?"symbol":typeof t},h=function(){function t(t,e){for(var n=0;n<e.length;n++){var o=e[n];o.enumerable=o.enumerable||!1,o.configurable=!0,"value"in o&&(o.writable=!0),Object.defineProperty(t,o.key,o)}}return function(e,n,o){return n&&t(e.prototype,n),o&&t(e,o),e}}(),p=function(t){function e(t,n){i(this,e);var o=a(this,(e.__proto__||Object.getPrototypeOf(e)).call(this));return o.resolveOptions(n),o.listenClick(t),o}return c(e,t),h(e,[{key:"resolveOptions",value:function(){var t=arguments.length>0&&void 0!==arguments[0]?arguments[0]:{};this.action="function"==typeof t.action?t.action:this.defaultAction,this.target="function"==typeof t.target?t.target:this.defaultTarget,this.text="function"==typeof t.text?t.text:this.defaultText,this.container="object"===d(t.container)?t.container:document.body}},{key:"listenClick",value:function(t){var e=this;this.listener=(0,f.default)(t,"click",function(t){return e.onClick(t)})}},{key:"onClick",value:function(t){var e=t.delegateTarget||t.currentTarget;this.clipboardAction&&(this.clipboardAction=null),this.clipboardAction=new l.default({action:this.action(e),target:this.target(e),text:this.text(e),container:this.container,trigger:e,emitter:this})}},{key:"defaultAction",value:function(t){return u("action",t)}},{key:"defaultTarget",value:function(t){var e=u("target",t);if(e)return document.querySelector(e)}},{key:"defaultText",value:function(t){return u("text",t)}},{key:"destroy",value:function(){this.listener.destroy(),this.clipboardAction&&(this.clipboardAction.destroy(),this.clipboardAction=null)}}],[{key:"isSupported",value:function(){var t=arguments.length>0&&void 0!==arguments[0]?arguments[0]:["copy","cut"],e="string"==typeof t?[t]:t,n=!!document.queryCommandSupported;return e.forEach(function(t){n=n&&!!document.queryCommandSupported(t)}),n}}]),e}(s.default);t.exports=p})},function(t,e){function n(t,e){for(;t&&t.nodeType!==o;){if("function"==typeof t.matches&&t.matches(e))return t;t=t.parentNode}}var o=9;if("undefined"!=typeof Element&&!Element.prototype.matches){var r=Element.prototype;r.matches=r.matchesSelector||r.mozMatchesSelector||r.msMatchesSelector||r.oMatchesSelector||r.webkitMatchesSelector}t.exports=n},function(t,e,n){function o(t,e,n,o,r){var a=i.apply(this,arguments);return t.addEventListener(n,a,r),{destroy:function(){t.removeEventListener(n,a,r)}}}function r(t,e,n,r,i){return"function"==typeof t.addEventListener?o.apply(null,arguments):"function"==typeof n?o.bind(null,document).apply(null,arguments):("string"==typeof t&&(t=document.querySelectorAll(t)),Array.prototype.map.call(t,function(t){return o(t,e,n,r,i)}))}function i(t,e,n,o){return function(n){n.delegateTarget=a(n.target,e),n.delegateTarget&&o.call(t,n)}}var a=n(4);t.exports=r},function(t,e){e.node=function(t){return void 0!==t&&t instanceof HTMLElement&&1===t.nodeType},e.nodeList=function(t){var n=Object.prototype.toString.call(t);return void 0!==t&&("[object NodeList]"===n||"[object HTMLCollection]"===n)&&"length"in t&&(0===t.length||e.node(t[0]))},e.string=function(t){return"string"==typeof t||t instanceof String},e.fn=function(t){return"[object Function]"===Object.prototype.toString.call(t)}},function(t,e){function n(t){var e;if("SELECT"===t.nodeName)t.focus(),e=t.value;else if("INPUT"===t.nodeName||"TEXTAREA"===t.nodeName){var n=t.hasAttribute("readonly");n||t.setAttribute("readonly",""),t.select(),t.setSelectionRange(0,t.value.length),n||t.removeAttribute("readonly"),e=t.value}else{t.hasAttribute("contenteditable")&&t.focus();var o=window.getSelection(),r=document.createRange();r.selectNodeContents(t),o.removeAllRanges(),o.addRange(r),e=o.toString()}return e}t.exports=n}])});

const messages = {
  'en': {
    'copy': 'Copy',
    'copy_to_clipboard': 'Copy to clipboard',
    'copy_success': 'Copied to clipboard!',
    'copy_failure': 'Failed to copy',
  },
  'es' : {
    'copy': 'Copiar',
    'copy_to_clipboard': 'Copiar al portapapeles',
    'copy_success': '¡Copiado!',
    'copy_failure': 'Error al copiar',
  },
  'de' : {
    'copy': 'Kopieren',
    'copy_to_clipboard': 'In die Zwischenablage kopieren',
    'copy_success': 'Kopiert!',
    'copy_failure': 'Fehler beim Kopieren',
  }
}

let locale = 'en'
if( document.documentElement.lang !== undefined
    && messages[document.documentElement.lang] !== undefined ) {
  locale = document.documentElement.lang
}

/**
 * Set up copy/paste for code blocks
 */

const runWhenDOMLoaded = cb => {
  if (document.readyState != 'loading') {
    cb()
  } else if (document.addEventListener) {
    document.addEventListener('DOMContentLoaded', cb)
  } else {
    document.attachEvent('onreadystatechange', function() {
      if (document.readyState == 'complete') cb()
    })
  }
}

const codeCellId = index => `codecell${index}`

// Clears selected text since ClipboardJS will select the text when copying
const clearSelection = () => {
  if (window.getSelection) {
    window.getSelection().removeAllRanges()
  } else if (document.selection) {
    document.selection.empty()
  }
}

// Changes tooltip text for two seconds, then changes it back
const temporarilyChangeTooltip = (el, newText) => {
  const oldText = el.getAttribute('data-tooltip')
  el.setAttribute('data-tooltip', newText)
  setTimeout(() => el.setAttribute('data-tooltip', oldText), 2000)
}

const addCopyButtonToCodeCells = () => {
  // If ClipboardJS hasn't loaded, wait a bit and try again. This
  // happens because we load ClipboardJS asynchronously.
  if (window.ClipboardJS === undefined) {
    setTimeout(addCopyButtonToCodeCells, 250)
    return
  }

  const codeCells = document.querySelectorAll('div.highlight-default:not(.output) pre, div.highlight-bash pre, div.highlight-sh pre, div.highlight-python pre')
  codeCells.forEach((codeCell, index) => {
    const id = codeCellId(index)
    codeCell.setAttribute('id', id)
    const pre_bg = getComputedStyle(codeCell).backgroundColor;

    const clipboardButton = id =>
    `<a class="copybtn o-tooltip--left" data-tooltip="${messages[locale]['copy']}" data-clipboard-target="#${id}">
      <img src="https://gitcdn.xyz/repo/choldgraf/sphinx-copybutton/master/sphinx_copybutton/_static/copy-button.svg" alt="${messages[locale]['copy_to_clipboard']}">
    </a>`
    codeCell.insertAdjacentHTML('afterend', clipboardButton(id))
  })

  const clipboard = new ClipboardJS('.copybtn')
  clipboard.on('success', event => {
    clearSelection()
    temporarilyChangeTooltip(event.trigger, messages[locale]['copy_success'])
  })

  clipboard.on('error', event => {
    temporarilyChangeTooltip(event.trigger, messages[locale]['copy_failure'])
  })
}

runWhenDOMLoaded(addCopyButtonToCodeCells);
"""

limit_output_length_css = r"""
pre.output {
  max-height: 400px;
  overflow-y: scroll;
}
div[class^="output"] {
    margin-bottom: 1em;
}
"""

tabbar_css = r"""
div.mdl-tabs__tab-bar { justify-content: left; }

.mdl-tabs .mdl-tabs__tab {
    font-size: 13px;
    height: 32px;
    line-height: 32px;
}

.mdl-tabs .mdl-tabs__tab-bar {
    height: 32px;
}

.mdl-tabs .mdl-tabs__tab-bar.text {
    margin-bottom: 12px;
}
"""

copybutton_css = r"""
a.copybtn {
    position: absolute;
    top: 12px;
    right: 24px;
    width: 1.2em;
    height: 1.2em;
    padding: 0 0 0 10px;
    opacity: .4;
    transition: opacity 0.5s;
    background-color: rgba(0, 0, 0, 0.02);
}

div.highlight  {
    position: relative;
}

.highlight:hover .copybtn {
	opacity: 1;
}

.o-tooltip--left {
    position: relative;
}

.o-tooltip--left:after {
    opacity: 0;
    visibility: hidden;
    position: absolute;
    content: attr(data-tooltip);
    padding: 2px 10px;
    top: -2px;
    left: 0;
    background: #666;
    font-size: 1rem;
    color: white;
    white-space: nowrap;
    z-index: 2;
    border-radius: 2px;
    transform: translateX(-102%) translateY(0);
    transition: opacity 0.2s cubic-bezier(0.64, 0.09, 0.08, 1), transform 0.2s cubic-bezier(0.64, 0.09, 0.08, 1);
}

.o-tooltip--left:hover:after {
    display: block;
    opacity: 1;
    visibility: visible;
    transform: translateX(-100%) translateY(0);
    transition: opacity 0.2s cubic-bezier(0.64, 0.09, 0.08, 1), transform 0.2s cubic-bezier(0.64, 0.09, 0.08, 1);
}
"""



"""Save codes into library"""
from typing import List
from d2lbook import notebook
import logging
import os

def _write_header(f):
    f.write('# This file is generated automatically through:\n')
    f.write('#    d2lbook build lib\n')
    f.write('# Don\'t edit it directly\n\n')

def save_mark(notebooks: List[str], lib_fname: str, save_mark: str):
    logging.info('Matching with the pattern: "%s"', save_mark)
    with open(lib_fname, 'w') as f:
        _write_header(f)
        lib_name = os.path.dirname(lib_fname)
        lib_name = lib_name.split('/')[-1]
        f.write('import sys\n'+lib_name+' = sys.modules[__name__]\n\n')

        for nb in notebooks:
            _save_code(nb, f, save_mark=save_mark)
        logging.info('Saved into %s', lib_fname)

def save_tab(notebooks: List[str], lib_fname: str, tab: str, default_tab: str):
    logging.info(f'Matching with the pattern: "#@save", seaching for tab {tab}')
    with open(lib_fname, 'w') as f:
        _write_header(f)
        for nb in notebooks:
           _save_code(nb, f, tab=tab, default_tab=default_tab)
        logging.info('Saved into %s', lib_fname)

def _save_code(input_fn, output_fp, save_mark=None, tab=None, default_tab=None):
    """get the code blocks (import, class, def) that will be saved"""
    with open(input_fn, 'r') as f:
        nb = notebook.read_markdown(f.read())
    if tab:
        nb = notebook.get_tab_notebook(nb, tab, default_tab)
        if not nb:
            return
    saved = []
    for cell in nb.cells:
        if cell.cell_type == 'code':
            lines = cell.source.split('\n')
            for i, l in enumerate(lines):
                if ((save_mark and l.strip().startswith('#') and save_mark in l) or
                    (tab and l.strip().endswith('@save') and '#'in l)):
                    if l.strip().startswith('#'):
                        block = [lines[i+1]]
                    else:
                        block = lines[i:i+2]
                    # For code block only containing import statements (e.g., in
                    # preface.md)
                    if lines[i+1].startswith('import') or lines[i+1].startswith('from'):
                        for j in range(i+2, len(lines)):
                            block.append(lines[j])
                    # For code blocks containing def or class
                    else:
                        for j in range(i+2, len(lines)):
                            l = lines[j]
                            if not l.startswith(' ') and len(l):
                                break
                            block.append(l)
                    if len(block[-1]) == 0:
                        del block[-1]
                    saved.append(block)

    if saved:
        logging.info('Found %d blocks in %s', len(saved), input_fn)
        for block in saved:
            logging.info(' --- %s', block[0])
            code = '# Defined in file: %s\n%s\n\n\n' %(input_fn, '\n'.join(block))
            output_fp.write(code)



import argparse
from d2lbook import markdown
import glob
import re
import sys

__all__  = ['activate']

commands = ['tab']

def activate():
    parser = argparse.ArgumentParser(description='Activate tabs')
    parser.add_argument('tab', default='all', help='the tab to activate')
    parser.add_argument('filename', help='the markdown files to activate')
    args = parser.parse_args(sys.argv[2:])

    for fn in glob.glob(args.filename):
        _activate_tab(fn, args.tab)

_tab_re = re.compile('# *@tab +([\w]+)')

def _get_cell_tab(cell):
    if cell['type'] != 'code':
        return None
    if not '.input' in cell['class'] and not 'python' in cell['class']:
        return None
    match = _tab_re.search(cell['source'])
    if match:
        return match[1]
    return 'default'

def _activate_tab(filename, tab):
    with open(filename, 'r') as f:
        src = f.read()
    cells = markdown.split_markdown(src)
    for cell in cells:
        cell_tab = _get_cell_tab(cell)
        if not cell_tab:
            continue
        if tab == 'all' or cell_tab == 'all' or cell_tab == tab:
            # activate
            cell['class'] = '{.python .input}'
        else: # disactivate
            cell['class'] = 'python'
    src = markdown.join_markdown_cells(cells)
    with open(filename, 'w') as f:
        f.write(src)



"""utilities to handle notebooks"""

from typing import Union, List, Optional

import os
import copy
import notedown
import nbformat
import nbconvert
from nbformat import notebooknode
from d2lbook import markdown
from d2lbook import common

def create_new_notebook(nb: notebooknode.NotebookNode,
                         cells: List[notebooknode.NotebookNode]
                         ) -> notebooknode.NotebookNode:
    """create an empty notebook by copying metadata from nb"""
    new_nb = copy.deepcopy(nb)
    new_nb.cells = cells
    return new_nb

def read(fname: str):
    if not os.path.exists(fname) or os.stat(fname).st_size == 0:
        return None
    with open(fname, 'r') as f:
        return nbformat.read(f, as_version=4)


def read_markdown(source: Union[str, List[str]]) -> notebooknode.NotebookNode:
    """Returns a notebook from markdown source"""
    if not isinstance(source, str):
        source = '\n'.join(source)
    reader = notedown.MarkdownReader(match='strict')
    return reader.reads(source)

def split_markdown_cell(nb: notebooknode.NotebookNode) -> notebooknode.NotebookNode:
    """split a markdown cell if it contains tab block.

    a new property `class` is added to the metadata for a tab cell.
    """
    # merge continous markdown cells
    grouped_cells = common.group_list(nb.cells,
                                      lambda cell, _: cell.cell_type=='markdown')
    new_cells = []
    for is_md, group in grouped_cells:
        if not is_md:
            new_cells.extend(group)
        else:
            src = '\n\n'.join(cell.source for cell in group)
            md_cells = markdown.split_markdown(src)
            is_tab_cell = lambda cell, _: cell['type']=='markdown' and 'class' in cell
            grouped_md_cells = common.group_list(md_cells, is_tab_cell)
            for is_tab, md_group in grouped_md_cells:
                new_cell = nbformat.v4.new_markdown_cell(
                    markdown.join_markdown_cells(md_group))
                if is_tab:
                    tab = md_group[0]['class']
                    assert tab.startswith('`') and tab.endswith('`'), tab
                    new_cell.metadata['tab'] = [tab[1:-1]]
                new_cells.append(new_cell)
    new_cells = [cell for cell in new_cells if cell.source]
    return create_new_notebook(nb, new_cells)

def _get_cell_tab(cell: notebooknode.NotebookNode, default_tab: str='') -> List[str]:
    """Get the cell tab"""
    if 'tab' in cell.metadata:
        return cell.metadata['tab']
    if cell.cell_type != 'code':
        return []
    match = common.source_tab_pattern.search(cell.source)
    if match:
        return [tab.strip() for tab in match[1].split(',')]
    return [default_tab,]

def get_tab_notebook(nb: notebooknode.NotebookNode, tab: str, default_tab: str
                     ) -> notebooknode.NotebookNode:
    """Returns a notebook with code/markdown cells that doesn't match tab
    removed.

    Return None if no cell matched the tab and nb contains code blocks.

    A `origin_pos` property is added to the metadata for each cell, which
    records its position in the original notebook `nb`.
    """
    matched_tab = False
    new_cells = []
    for i, cell in enumerate(nb.cells):
        new_cell = copy.deepcopy(cell)
        new_cell.metadata['origin_pos'] = i
        cell_tab = _get_cell_tab(new_cell, default_tab)
        if not cell_tab:
            new_cells.append(new_cell)
        else:
            if cell_tab == ['all'] or tab in cell_tab:
                new_cell.metadata['tab'] = 'all' if cell_tab == ['all'] else tab
                matched_tab = True
                # remove the tab from source
                lines = new_cell.source.split('\n')
                for j, line in enumerate(lines):
                    src_tab = common.source_tab_pattern.search(line)
                    text_tab = common.md_mark_pattern.search(line)
                    if src_tab or (text_tab and (
                            text_tab[1]=='begin_tab' or text_tab[1]=='end_tab')):
                        del lines[j]
                    new_cell.source = '\n'.join(lines)
                new_cells.append(new_cell)
    if not matched_tab and any([cell.cell_type=='code' for cell in nb.cells]):
        return None
    return create_new_notebook(nb, new_cells)

def merge_tab_notebooks(src_notebooks: List[notebooknode.NotebookNode]
                        ) -> notebooknode.NotebookNode:
    """Merge the tab notebooks into a single one.

    The reserved function of get_tab_notebook.
    """
    n = max([max([cell.metadata['origin_pos'] for cell in nb.cells])
             for nb in src_notebooks])
    new_cells = [None] * (n+1)
    for nb in src_notebooks:
        for cell in nb.cells:
            new_cells[cell.metadata['origin_pos']] = copy.deepcopy(cell)
    return create_new_notebook(src_notebooks[0], new_cells)

def _get_tab_bar(tabs, tab_id, default_tab, div_class=''):
    code = f"```eval_rst\n\n.. raw:: html\n\n    <div class=\"mdl-tabs mdl-js-tabs mdl-js-ripple-effect\"><div class=\"mdl-tabs__tab-bar {div_class}\">"
    for i, tab in enumerate(tabs):
        active = 'is-active' if tab == default_tab else ''
        code +=f'<a href="#{tab}-{tab_id}-{i}" onclick="tagClick(\'{tab}\'); return false;" class="mdl-tabs__tab {active}">{tab}</a>'
    code += "</div>\n```"
    return nbformat.v4.new_markdown_cell(code)

def _get_tab_panel(cells, tab, tab_id, default_tab):
    active = 'is-active' if tab == default_tab else ''
    tab_panel_begin = nbformat.v4.new_markdown_cell(
        f"```eval_rst\n.. raw:: html\n\n    <div class=\"mdl-tabs__panel {active}\" id=\"{tab}-{tab_id}\">\n```")
    tab_panel_end = nbformat.v4.new_markdown_cell(
        "```eval_rst\n.. raw:: html\n\n    </div>\n```")
    return [tab_panel_begin, *cells, tab_panel_end]


def _merge_tabs(nb: notebooknode.NotebookNode):
    """merge side-by-side tabs into a single one"""
    def _tab_status(cell, status):
        tab = _get_cell_tab(cell)
        if tab:
            if cell.cell_type == 'markdown':
                return 1
            if tab == 'all':
                return 0
            return 2
        return 0
    cell_groups = common.group_list(nb.cells, _tab_status)
    meta = [(in_tab, [cell.metadata['tab'] for cell in group] if in_tab else None
             ) for in_tab, group in cell_groups]
    new_cells = []
    i = 0
    while i < len(meta):
        in_tab, tabs = meta[i]
        if not in_tab:
            new_cells.append((False, cell_groups[i][1]))
            i += 1
        else:
            j = i + 1
            while j < len(meta):
                if meta[j][1] != tabs:
                    break
                j += 1
            groups = [group for _, group in cell_groups[i:j]]
            new_cells.append((True, [x for x in zip(*groups)]))
            i = j

    return new_cells

def add_html_tab(nb: notebooknode.NotebookNode, default_tab: str) -> notebooknode.NotebookNode:
    """Add html codes for the tabs"""
    cell_groups = _merge_tabs(nb)
    tabs = [len(group) for in_tab, group in cell_groups if in_tab]
    if not tabs or max(tabs) <= 1:
        return nb
    new_cells = []
    for i, (in_tab, group) in enumerate(cell_groups):
        if not in_tab:
            new_cells.extend(group)
        else:
            tabs = [cells[0].metadata['tab'] for cells in group]
            div_class = "code" if group[0][0].cell_type == 'code' == 2 else "text"
            new_cells.append(_get_tab_bar(tabs, i, default_tab, div_class))
            for j, (tab, cells) in enumerate(zip(tabs, group)):
                new_cells.extend(_get_tab_panel(cells, tab, f'{i}-{j}', default_tab))
            new_cells.append(nbformat.v4.new_markdown_cell(
                "```eval_rst\n.. raw:: html\n\n    </div>\n```"))
    return create_new_notebook(nb, new_cells)



import os
import sys
import logging
import argparse
import shutil
from d2lbook.utils import *
from d2lbook.config import Config
from d2lbook import colab
from d2lbook import sagemaker

__all__  = ['deploy']

commands = ['html', 'pdf', 'pkg', 'colab', 'sagemaker', 'all']

def deploy():
    parser = argparse.ArgumentParser(description='Deploy documents')
    parser.add_argument('commands', nargs='+', choices=commands)
    args = parser.parse_args(sys.argv[2:])
    config = Config()
    if config.deploy['s3_bucket']:
        deployer = S3Deployer(config)
    elif config.deploy['github_repo']:
        deployer = GithubDeployer(config)
    else:
        logging.fatal('No deployment URL. You need to specify either'
                      'a Github repo or a S3 bucket')
        exit(-1)
    for cmd in args.commands:
        getattr(deployer, cmd)()

class Deployer(object):
    def __init__(self, config):
        self.config = config

    def colab(self):
        _colab = colab.Colab(self.config)
        if not _colab.valid():
            return
        def _run():
            repo = _colab.git_repo(self.config.tab)
            bash_fname = os.path.join(os.path.dirname(__file__), 'upload_github.sh')
            run_cmd(['bash', bash_fname, self.config.colab_dir, repo])
        tab = self.config.tab
        self.config.set_tab('all')
        self.config.iter_tab(_run)
        self.config.set_tab(tab)

    def sagemaker(self):
        _sagemaker = sagemaker.Sagemaker(self.config)
        if not _sagemaker.valid():
            return
        def _run():
            repo = _sagemaker.git_repo(self.config.tab)
            bash_fname = os.path.join(os.path.dirname(__file__), 'upload_github.sh')
            run_cmd(['bash', bash_fname, self.config.sagemaker_dir, repo])
        tab = self.config.tab
        self.config.set_tab('all')
        self.config.iter_tab(_run)
        self.config.set_tab(tab)

class GithubDeployer(Deployer):
    def __init__(self, config):
        super(GithubDeployer, self).__init__(config)
        self.git_dir = os.path.join(self.config.tgt_dir, 'github_deploy')
        shutil.rmtree(self.git_dir, ignore_errors=True)
        mkdir(self.git_dir)

    def html(self):
        run_cmd(['cp -r', os.path.join(self.config.html_dir, '*'), self.git_dir])

    def pdf(self):
        shutil.copy(self.config.pdf_fname, self.git_dir)

    def pkg(self):
        shutil.copy(self.config.pkg_fname, self.git_dir)

    def __del__(self):
        bash_fname = os.path.join(os.path.dirname(__file__), 'upload_github.sh')
        run_cmd(['bash', bash_fname, self.git_dir, self.config.deploy['github_repo']])

class S3Deployer(Deployer):
    def __init__(self, config):
        super(S3Deployer, self).__init__(config)

    def html(self):
        bash_fname = os.path.join(os.path.dirname(__file__), 'upload_doc_s3.sh')
        run_cmd(['bash', bash_fname, self.config.html_dir, self.config.deploy['s3_bucket']])
        self.colab()
        self.sagemaker()

    def pdf(self):
        url = self.config.deploy['s3_bucket']
        if not url.endswith('/'):
            url += '/'
        logging.info('cp %s to %s', self.config.pdf_fname, url)
        run_cmd(['aws s3 cp', self.config.pdf_fname, url, "--acl 'public-read' --quiet"])

    def _deploy_other_files(self, tgt_url):
        other_urls = self.config.deploy['other_file_s3urls'].split()
        for other_url in other_urls:
            logging.info('cp %s to %s', other_url, tgt_url)
            run_cmd(['aws s3 cp', other_url, tgt_url, "--acl 'public-read' --quiet"])

    def pkg(self):
        url = self.config.deploy['s3_bucket']
        if not url.endswith('/'):
            url += '/'
        logging.info('cp %s to %s', self.config.pkg_fname, url)
        run_cmd(['aws s3 cp', self.config.pkg_fname, url, "--acl 'public-read' --quiet"])
        self._deploy_other_files(url)

    def all(self):
        self.html()
        self.pdf()
        self.pkg()



"""utilities to handle markdown
"""
from d2lbook import common
from typing import List, Dict

def split_markdown(source: str) -> List[Dict[str, str]]:
    """Split markdown into a list of text and code cells.

    A cell has three fields:
    1. type: either code or markdown
    2. class: code class or tab class
    3. source: single string for the source
    """
    cells: List[Dict] = []
    in_code = False
    in_tab = False
    cur_code_mark = None
    cur_tag = None
    cur_src = []
    def _add_cell(cur_src: List[str], cells: List[Dict]):
        if cur_src:
            src = '\n'.join(cur_src)
            if in_code:
                cells.append({'type':'code', 'fence':cur_code_mark, 'class':cur_tag, 'source':src})
            else:
                cells.append({'type':'markdown', 'source':src})
                if cur_tag:
                    cells[-1]['class'] = cur_tag

    for l in source.split('\n'):
        code = common.md_code_fence.match(l)
        tab = common.md_mark_pattern.match(l)
        if code:
            # code can be nested
            if in_tab or (in_code and code.groups()[0] != cur_code_mark):
                cur_src.append(l)
            else:
                _add_cell(cur_src, cells)
                cur_src = []
                cur_code_mark, cur_tag = code.groups()
                in_code ^= True
        elif tab:
            begin = tab.groups()[0] == 'begin_tab'
            end = tab.groups()[0] == 'end_tab'
            if in_code or (not begin and not end):
                cur_src.append(l)
            else:
                _add_cell(cur_src, cells)
                cur_src = []
                if begin:
                    cur_tag = tab.groups()[1]
                else:
                    cur_tag = None
                in_tab = begin
        else:
            cur_src.append(l)
    _add_cell(cur_src, cells)
    return cells

def join_markdown_cells(cells: List[Dict]) -> str:
    """Join a list of cells into a markdown string"""
    src = []
    for c in cells:
        if c['type'] == 'markdown':
            if 'class' in c:
                src.append(f':begin_tab:{c["class"]}')
            src.append(c['source'])
            if 'class' in c:
                src.append(':end_tab:')
        else:
            src += [c['fence']+c['class'], c['source'], c['fence']]
    return '\n'.join(src)



"""Integration with Colab notebooks"""
import os
import re
import nbformat
import logging
from d2lbook import notebook
from d2lbook import utils

def parse_repo_lib(repo_str, lib_str, version):
    repo = utils.split_config_str(repo_str)
    if len(repo) == 1 and len(repo[0]) == 1:
        repos = {None:repo[0]}
        libs = {None:utils.split_config_str(lib_str, 2)}
    else:
        repo = utils.split_config_str(repo_str, 2)
        repos = {r[0]:r[1] for r in repo}
        libs_list = utils.split_config_str(lib_str, 3)
        libs = {}
        for tab, pkg, install in libs_list:
            if tab in libs:
                libs[tab].append([pkg, install])
            else:
                libs[tab] = [[pkg, install]]
    for tab in libs:
        for i, l in enumerate(libs[tab]):
            if '==version' in l[1]:
                libs[tab][i][1] = l[1].replace('==version', f'=={version}')
    return repos, libs


class Colab():
    def __init__(self, config):
        self._valid = config.colab and config.colab['github_repo']
        if not self._valid:
            return
        self.tabs = config.tabs
        self.config = config.colab
        self._repo, self._libs = parse_repo_lib(
            self.config['github_repo'], self.config['libs'], config.library["version"])

    def valid(self):
        return self._valid

    def git_repo(self, tab):
        return self._repo[tab]

    def generate_notebooks(self, eval_dir, colab_dir, tab):
        if not self._valid:
            return
        # copy notebook fron eval_dir to colab_dir
        utils.run_cmd(['rm -rf', colab_dir])
        utils.run_cmd(['cp -r', eval_dir, colab_dir])
        notebooks = utils.find_files('**/*.ipynb', colab_dir)
        for fn in notebooks:
            nb = notebook.read(fn)
            if not nb:
                continue
            # Use Python3 as the kernel
            update_notebook_kernel(nb, "python3", "Python 3")
            # Check if GPU is needed
            use_gpu = False
            for cell in nb.cells:
                if cell.cell_type == 'code':
                    if self.config['gpu_pattern'] in cell.source:
                        use_gpu = True
                        break
            if use_gpu:
                nb['metadata'].update({"accelerator": "GPU"})
                logging.info('Use GPU for '+fn)
            # Update SVG image URLs
            if self.config['replace_svg_url']:
                _update_svg_urls(nb, self.config['replace_svg_url'], fn, colab_dir)
            insert_additional_installation(nb, self._libs[tab], self.config['libs_header'])
            with open(fn, 'w') as f:
                f.write(nbformat.writes(nb))

    def add_button(self, html_dir):
        """Add an open colab button in HTML"""
        if not self._valid:
            return
        files = utils.find_files('**/*.html', html_dir, self.config['exclusions'])
        for fn in files:
            with open(fn, 'r') as f:
                html = f.read()
            if 'id="Colab' in html:
                continue
            url = os.path.relpath(fn, html_dir).replace('.html', '.ipynb')
            if self.tabs:
                colab_html = ''
                for tab in self.tabs:
                    colab_tab = _get_colab_html(self._repo[tab], url, f'Colab [{tab}]')
                    colab_html += f'<div class="d2l-tabs__tab">{colab_tab}</div>'
                colab_html = f'<div class="d2l-tabs" style="float:right">{colab_html}</div>'
            else:
                colab_html = _get_colab_html(self._repo[None], url, f'Colab')
            html = html.replace('</h1>', colab_html+'</h1>')
            with open(fn, 'w') as f:
                f.write(html)

def _get_colab_html(repo, url, text):
    id = text.replace(" ", "_")
    colab_link = f'https://colab.research.google.com/github/{repo}/blob/master/{url}'
    colab_html = f'<a href="{colab_link}" onclick="captureOutboundLink(\'{colab_link}\'); return false;"> <button style="float:right", id="{id}" class="mdl-button mdl-js-button mdl-button--primary mdl-js-ripple-effect"> <i class=" fas fa-external-link-alt"></i> {text} </button></a><div class="mdl-tooltip" data-mdl-for="{id}"> Open the notebook in Colab</div>'
    return colab_html

def insert_additional_installation(notebook, lib, lib_header):
    if lib:
        cell = _get_installation_cell(notebook, lib)
        if cell:
            notebook.cells.insert(0, cell)
            if lib_header:
                notebook.cells.insert(
                    0, nbformat.v4.new_markdown_cell(source=lib_header))

def update_notebook_kernel(notebook, name, display_name=None):
    if not display_name:
        display_name = name
    notebook['metadata'].update({"kernelspec": {
        "name": name,
        "display_name": display_name
    }})


def _update_svg_urls(notebook, pattern, filename, root_dir):
    orgin_url, new_url = utils.split_config_str(pattern, 2)[0]
    svg_re = re.compile('!\[.*\]\(([\.-_\w\d]+\.svg)\)')
    for cell in notebook.cells:
        if cell.cell_type == 'markdown':
            lines = cell.source.split('\n')
            for i, l in enumerate(lines):
                m = svg_re.search(l)
                if not m:
                    continue
                path = os.path.relpath(os.path.realpath(os.path.join(
                    root_dir, os.path.basename(filename), m[1])), root_dir)
                if not path.startswith(orgin_url):
                    logging.warning("%s in %s does not start with %s"
                                    "specified by replace_svg_url"%(
                                        path, filename, orgin_url))
                else:
                    url = new_url + path[len(orgin_url):]
                    lines[i] = l.replace(m[1], url)
            cell.source = '\n'.join(lines)

def _get_installation_cell(notebook, libs):
    """Return a cell for installing the additional libs"""
    lib_dict = dict(libs)
    lib1_re = re.compile('from ([_\w\d]+) import')
    lib2_re = re.compile('import ([_\w\d]+)')
    find_libs = []
    for cell in notebook.cells:
        if cell.cell_type == 'code':
            lines = cell.source.split('\n')
            for l in lines:
                if l.strip().startswith('#'): # it's a comment
                    continue
                m = lib1_re.search(l)
                if not m:
                    m = lib2_re.search(l)
                if m and m[1] in lib_dict:
                    find_libs.append(m[1])
    if not find_libs:
        return None
    install_str = ''
    for lib in set(find_libs):
        install_str += '!pip install ' + lib_dict[lib] + '\n'
    return nbformat.v4.new_code_cell(source=install_str)



import argparse
import sys
from d2lbook.build import build
from d2lbook.deploy import deploy
from d2lbook.clear import clear
from d2lbook.activate import activate
import logging

logging.basicConfig(format='[d2lbook:%(filename)s:L%(lineno)d] %(levelname)-6s %(message)s')
logging.getLogger().setLevel(logging.INFO)


def main():
    commands = {'build': build, 'deploy':deploy, 'clear':clear, 'activate':activate}
    parser = argparse.ArgumentParser(description='''
D2L Book: Publish a book based on Jupyter notebooks.

Run d2lbook command -h to get the help message for each command.
''')
    parser.add_argument('command', nargs=1, choices=commands.keys())
    args = parser.parse_args(sys.argv[1:2])
    commands[args.command[0]]()

if __name__ == "__main__":
    main()



import os
import sys
import notedown
import nbformat
import logging
import shutil
import datetime
import argparse
import re
import regex
import subprocess
import hashlib
from d2lbook.utils import *  # TODO(mli), don't report *
from d2lbook.sphinx import prepare_sphinx_env
from d2lbook.config import Config
from d2lbook import colab, sagemaker
from d2lbook import markdown
from d2lbook import library
from d2lbook import notebook
from d2lbook import rst as rst_lib

__all__  = ['build']

commands = ['eval', 'rst', 'html', 'pdf', 'pkg', 'linkcheck', 'ipynb',
            'outputcheck', 'lib', 'colab', 'sagemaker', 'all', 'merge']

def build():
    parser = argparse.ArgumentParser(description='Build the documents')
    parser.add_argument('commands', nargs='+', choices=commands)
    parser.add_argument('--tab', default=None, help='The tab to build, if multi-tab is enabled.')
    args = parser.parse_args(sys.argv[2:])
    config = Config(tab=args.tab)
    builder = Builder(config)
    for cmd in args.commands:
        getattr(builder, cmd)()

def _once(func):
    # An decorator that run a method only once
    def warp(self):
        name = func.__name__
        if self.config.tab:
            name += '_' + self.config.tab
        if name in self.done and self.done[name]:
            return
        full_name = 'd2lbook build ' + name
        tik = datetime.datetime.now()
        func(self)
        logging.info('=== Finished "%s" in %s', full_name,
                        get_time_diff(tik, datetime.datetime.now()))
        self.done[name] = True
    return warp

class Builder(object):
    def __init__(self, config):
        self.config = config
        mkdir(self.config.tgt_dir)
        self.sphinx_opts = '-j 4'
        if config.build['warning_is_error'].lower() == 'true':
            self.sphinx_opts += ' -W'
        self.done = {}
        self._colab = colab.Colab(config)
        self._sagemaker = sagemaker.Sagemaker(config)

    def _find_md_files(self):
        build = self.config.build
        src_dir = self.config.src_dir
        notebooks = find_files(build['notebooks'], src_dir,
                               build['exclusions']+' '+build['non-notebooks'])
        pure_markdowns = find_files(build['non-notebooks'], src_dir,
                                    build['exclusions'])
        depends = find_files(build['dependencies'], src_dir)
        return sorted(notebooks), sorted(pure_markdowns), sorted(depends)

    def _get_updated_md_files(self):
        notebooks, pure_markdowns, depends = self._find_md_files()
        depends_mtimes = get_mtimes(depends)
        latest_depend = max(depends_mtimes) if len(depends_mtimes) else 0
        updated_notebooks = get_updated_files(
            notebooks, self.config.src_dir, self.config.eval_dir, 'md', 'ipynb', latest_depend)
        updated_markdowns = get_updated_files(
            pure_markdowns, self.config.src_dir, self.config.eval_dir, 'md', 'md', latest_depend)
        return updated_notebooks, updated_markdowns

    def outputcheck(self):
        notebooks, _, _ = self._find_md_files()
        reader = notedown.MarkdownReader()
        error = False
        for fn in notebooks:
            with open(fn, 'r') as f:
                notebook= reader.read(f)
            for c in notebook.cells:
                if 'outputs' in c and len(c['outputs']):
                    logging.error("Found execution outputs in %s", fn)
                    error = True
        if error:
            exit(-1)

    @_once
    def eval(self):
        """Evaluate the notebooks and save them in a different folder"""
        # TODO(mli) if tabs is enabled, and a .md doesn't have the default tab,
        # then the current implementation will not run the eval.
        eval_tik = datetime.datetime.now()
        notebooks, pure_markdowns, depends = self._find_md_files()
        depends_mtimes = get_mtimes(depends)
        latest_depend = max(depends_mtimes) if len(depends_mtimes) else 0
        updated_notebooks = get_updated_files(
            notebooks, self.config.src_dir, self.config.eval_dir, 'md', 'ipynb', latest_depend)
        updated_markdowns = get_updated_files(
            pure_markdowns, self.config.src_dir, self.config.eval_dir, 'md', 'md', latest_depend)
        num_updated_notebooks = len(updated_notebooks)
        num_updated_markdowns = len(updated_markdowns)
        logging.info('%d notebooks are outdated', num_updated_notebooks)
        for i, nb in enumerate(updated_notebooks):
            logging.info('[%d] %s', i + 1, nb[0])
        self._copy_resources(self.config.src_dir, self.config.eval_dir)

        for i, (src, tgt) in enumerate(updated_notebooks):
            tik = datetime.datetime.now()
            logging.info('[%d/%d, %s] Evaluating %s, save as %s',
                         i+1, num_updated_notebooks,
                         get_time_diff(eval_tik, tik), src, tgt)
            mkdir(os.path.dirname(tgt))
            run_cells = self.config.build['eval_notebook'].lower()=='true'
            process_and_eval_notebook(src, tgt, run_cells, tab=self.config.tab,
                                      default_tab=self.config.default_tab)
            tok = datetime.datetime.now()
            logging.info('Finished in %s', get_time_diff(tik, tok))

        for src, tgt in updated_markdowns:
            logging.info('Copying %s to %s', src, tgt)
            mkdir(os.path.dirname(tgt))
            shutil.copyfile(src, tgt)
        self._rm_tgt_files('md', 'ipynb', self.config.eval_dir)

    # Remove target files (e.g., eval and rst) based on removed files under src
    def _rm_tgt_files(self, src_ext, tgt_ext, tgt_dir, must_incls=None):
        notebooks_to_rm = get_files_to_rm(
                self.config.build['notebooks'], self.config.src_dir, tgt_dir,
                src_ext, tgt_ext)

        non_notebooks_pattern = (self.config.build['non-notebooks'] + ' '
                + self.config.build['resources'])
        non_notebooks_to_rm = get_files_to_rm(non_notebooks_pattern,
                self.config.src_dir, tgt_dir)

        if must_incls:
            must_incls = set(must_incls)
        tgt_files_to_rm = [f for f in notebooks_to_rm + non_notebooks_to_rm
                           if not must_incls or f not in must_incls]
        if tgt_files_to_rm:
            tgt_files_to_rm_concise = hide_individual_data_files(tgt_files_to_rm)
            logging.info('Cleaning target files whose corresponding source '
                'files are removed (individual data files are hidden, '
                'e.g., _build/eval/data/VOC/A/B/C -> _build/eval/data/VOC')
            for fn in tgt_files_to_rm_concise:
                logging.info('Cleaning target files: %s' % fn)
            for fn in tgt_files_to_rm:
                os.remove(fn)
            rmed_empty_dirs = []  # To display more concisely
            rm_empty_dir(tgt_dir, rmed_empty_dirs)
            rmed_empty_dirs_concise = hide_individual_data_files(rmed_empty_dirs)
            for fn in rmed_empty_dirs_concise:
                logging.info('Cleaned empty directories: %s' % fn)

    def _copy_resources(self, src_dir, tgt_dir):
        resources = self.config.build['resources']
        if resources:
            logging.info('Copying resources "%s" from %s to %s',
                         ' '.join(resources.split()), src_dir, tgt_dir)
        for res in resources.split():
            src = os.path.join(src_dir, res)
            updated = get_updated_files(find_files(src), src_dir, tgt_dir)
            for src, tgt in updated:
                if os.path.isdir(src):
                    continue
                mkdir(os.path.dirname(tgt))
                shutil.copyfile(src, tgt)

    def _copy_rst(self):
        rst_files = find_files(self.config.build['rsts'], self.config.src_dir)
        updated_rst = get_updated_files(rst_files, self.config.src_dir, self.config.rst_dir)
        if len(updated_rst):
            logging.info('Copy %d updated RST files to %s',
                         len(updated_rst), self.config.rst_dir)
        for src, tgt in updated_rst:
            copy(src, tgt)
        return rst_files

    @_once
    def merge(self):
        assert self.config.tab == 'all'
        assert self.config.eval_dir.endswith('_all')
        assert len(self.config.tabs) > 1, self.config.tabs
        default_eval_dir = self.config.eval_dir[:-4]
        notebooks = find_files(os.path.join(default_eval_dir, '**', '*.ipynb'))
        # TODO(mli) if no default tab, then will not trigger merge
        updated_notebooks = get_updated_files(
            notebooks, default_eval_dir, self.config.eval_dir, 'ipynb', 'ipynb')
        tab_dirs = [default_eval_dir+'_'+tab for tab in self.config.tabs[1:]]
        for default, merged in updated_notebooks:
            src_notebooks = [default]
            for tab_dir in tab_dirs:
                fname = os.path.join(tab_dir,
                    os.path.relpath(default, default_eval_dir))
                if os.path.exists(fname) and os.stat(fname).st_size > 0:
                    src_notebooks.append(fname)
            logging.info(f'merge {src_notebooks} into {merged}')
            src_nbs = [nbformat.read(open(fn, 'r'), as_version=4)
                       for fn in src_notebooks]
            if len(src_nbs) > 1:
                dst_nb = notebook.merge_tab_notebooks(src_nbs)
                dst_nb = notebook.add_html_tab(dst_nb, self.config.default_tab)
            else:
                dst_nb = src_nbs[0]
            mkdir(os.path.dirname(merged))
            with open(merged, 'w') as f:
                nbformat.write(dst_nb, f)
        self._copy_resources(default_eval_dir, self.config.eval_dir)

    @_once
    def rst(self):
        if self.config.tab == 'all':
            self.merge()
        else:
            self.eval()
        notebooks = find_files(os.path.join(self.config.eval_dir, '**', '*.ipynb'))
        updated_notebooks = get_updated_files(
            notebooks, self.config.eval_dir, self.config.rst_dir, 'ipynb', 'rst')
        logging.info('%d rst files are outdated', len(updated_notebooks))
        for src, tgt in updated_notebooks:
            logging.info('Convert %s to %s', src, tgt)
            mkdir(os.path.dirname(tgt))
            ipynb2rst(src, tgt)
        # Generate conf.py under rst folder
        prepare_sphinx_env(self.config)
        self._copy_rst()
        self._copy_resources(self.config.src_dir, self.config.rst_dir)

        must_incl_rst_files = get_tgt_files_from_src_pattern(
                self.config.build['rsts'], self.config.rst_dir, 'rst', 'rst')
        self._rm_tgt_files('md', 'rst', self.config.rst_dir,
                           must_incl_rst_files)

    @_once
    def html(self):
        self.rst()
        self.colab()
        self.sagemaker()
        run_cmd(['sphinx-build', self.config.rst_dir, self.config.html_dir,
                 '-b html -c', self.config.rst_dir, self.sphinx_opts])
        self._colab.add_button(self.config.html_dir)

    def _default_tab_dir(self, dirname):
        tokens = dirname.split('/')
        if self.config.tabs and '_' in tokens[-1]:
            tokens[-1] =  '_'.join(tokens[-1].split('_')[:-1])
            return '/'.join(tokens)
        return dirname

    @_once
    def ipynb(self):
        self.eval()
        run_cmd(['rm -rf', self.config.ipynb_dir, '; cp -r ',
                 self.config.eval_dir, self.config.ipynb_dir])
        update_ipynb_toc(self.config.ipynb_dir)

    @_once
    def colab(self):
        def _run():
            self.ipynb()
            self._colab.generate_notebooks(self.config.ipynb_dir,
                                           self.config.colab_dir, self.config.tab)
        self.config.iter_tab(_run)

    @_once
    def sagemaker(self):
        def _run():
            self.ipynb()
            self._sagemaker.generate_notebooks(self.config.ipynb_dir,
                                               self.config.sagemaker_dir, self.config.tab)
        self.config.iter_tab(_run)

    @_once
    def linkcheck(self):
        self.rst()
        run_cmd(['sphinx-build', self.config.rst_dir, self.config.linkcheck_dir,
                 '-b linkcheck -c', self.config.rst_dir, self.sphinx_opts])

    @_once
    def pdf(self):
        self.rst()
        run_cmd(['sphinx-build ', self.config.rst_dir, self.config.pdf_dir,
                 '-b latex -c', self.config.rst_dir, self.sphinx_opts])

        script = self.config.pdf['post_latex']
        process_latex(self.config.tex_fname, script)
        run_cmd(['cd', self.config.pdf_dir, '&& make'])

    @_once
    def pkg(self):
        zip_fname = 'out.zip'
        if not self.config.tabs:
            self.ipynb()
            run_cmd(['cd', self.config.ipynb_dir, '&& zip -r',
                     zip_fname, '*'])
            shutil.move(os.path.join(self.config.ipynb_dir, zip_fname),
                        self.config.pkg_fname)
        else:
            origin_tab = self.config.tab
            for tab in self.config.tabs:
                self.config.set_tab(tab)
                self.ipynb()
                run_cmd(['rm -rf', tab])
                run_cmd(['cp -r', self.config.ipynb_dir, tab])
            run_cmd(['zip -r', zip_fname]+self.config.tabs)
            shutil.move(zip_fname, self.config.pkg_fname)
            self.config.set_tab(origin_tab)

    @_once
    def lib(self):
        root = os.path.join(self.config.src_dir, self.config.build['index'] + '.md')
        notebooks = get_toc(root)
        notebooks_enabled, _, _ = self._find_md_files()
        notebooks = [nb for nb in notebooks if nb in notebooks_enabled]

        save_patterns = self.config.library['save_patterns']
        if save_patterns:
            items = split_config_str(save_patterns, num_items_per_line=2)
            for lib_fname, tab in items:
               library.save_tab(notebooks, lib_fname, tab, self.config.default_tab)

        save_mark = self.config.library['save_mark']
        lib_fname = self.config.library['save_filename']
        if save_mark and lib_fname:
            library.save_mark(notebooks, lib_fname, save_mark)

        version = self.config.library['version']
        version_fn = self.config.library['version_file']
        if version and version_fn:
            with open(version_fn, 'r') as f:
                lines = f.read().split('\n')
            for i, l in enumerate(lines):
                if '__version__' in l:
                    lines[i] = f'__version__ = "{version}"'
                    logging.info(f'save {lines[i]} into {version_fn}')
            with open(version_fn, 'w') as f:
                f.write('\n'.join(lines))

    def all(self):
        self.eval()
        self.rst()
        self.html()
        self.pdf()
        self.pkg()

def update_ipynb_toc(root):
    """Change the toc code block into a list of clickable links"""
    notebooks = find_files('**/*.ipynb', root)
    for fn in notebooks:
        nb = notebook.read(fn)
        if not nb:
            continue
        for cell in nb.cells:
            if (cell.cell_type == 'markdown' and '```toc' in cell.source):
                md_cells = markdown.split_markdown(cell.source)
                for c in md_cells:
                    if c['type'] == 'code' and c['class'] == 'toc':
                        toc = []
                        for l in c['source'].split('\n'):
                            if l and not l.startswith(':'):
                                toc.append(' - [%s](%s.ipynb)'%(l,l))
                        c['source'] = '\n'.join(toc)
                        c['type'] = 'markdown'
                cell.source = markdown.join_markdown_cells(md_cells)
        with open(fn, 'w') as f:
            f.write(nbformat.writes(nb))

def get_toc(root):
    """return a list of files in the order defined by TOC"""
    subpages = get_subpages(root)
    res = [root]
    for fn in subpages:
        res.extend(get_toc(fn))
    return res

def get_subpages(input_fn):
    """read toc in input_fn, returns what it contains"""
    subpages = []
    reader = notedown.MarkdownReader()
    with open(input_fn, 'r') as f:
        nb = reader.read(f)
    for cell in nb.cells:
        if (cell.cell_type == 'code' and
            'attributes' in cell.metadata and
            'toc' in cell.metadata.attributes['classes']):
            for l in cell.source.split('\n'):
                l = l.strip()
                if not l.startswith(':'):
                    fn = os.path.join(os.path.dirname(input_fn), l + '.md')
                    if os.path.exists(fn):
                        subpages.append(fn)
    return subpages

def process_and_eval_notebook(input_fn, output_fn, run_cells, timeout=20*60,
                              lang='python', tab=None, default_tab=None):
    with open(input_fn, 'r') as f:
        md = f.read()
    nb = notebook.read_markdown(md)
    if tab:
        # get the tab
        nb = notebook.split_markdown_cell(nb)
        nb = notebook.get_tab_notebook(nb, tab, default_tab)
        if not nb:
            logging.info(f"Skip to eval tab {tab} for {input_fn}")
            # write an emtpy file to track the dependencies
            open(output_fn, 'w')
            return
    # evaluate
    if run_cells:
        # change to the notebook directory to resolve the relpaths properly
        cwd = os.getcwd()
        os.chdir(os.path.join(cwd, os.path.dirname(output_fn)))
        notedown.run(nb, timeout)
        os.chdir(cwd)
    # write
    nb['metadata'].update({'language_info':{'name':lang}})
    with open(output_fn, 'w') as f:
        f.write(nbformat.writes(nb))


def ipynb2rst(input_fn, output_fn):
    with open(input_fn, 'r') as f:
        nb = nbformat.read(f, as_version=4)
    sig = hashlib.sha1(input_fn.encode()).hexdigest()[:6]
    resources = {'unique_key':
                 'output_'+rm_ext(os.path.basename(output_fn))+'_'+sig}
    body, resources = rst_lib.convert_notebook(nb, resources)
    with open(output_fn, 'w') as f:
        f.write(body)
    outputs = resources['outputs']
    base_dir = os.path.dirname(output_fn)
    for fn in outputs:
        full_fn = os.path.join(base_dir, fn)
        with open(full_fn, 'wb') as f:
            f.write(outputs[fn])

def process_latex(fname, script):
    with open(fname, 'r') as f:
        lines = f.read().split('\n')

    _combine_citations(lines)
    _center_graphics(lines)

    with open(fname, 'w') as f:
        f.write('\n'.join(lines))
    # Execute custom process_latex script
    if script:
        cmd = "python " + script + " " + fname
        process = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE,
                                   stderr=subprocess.PIPE)
        stdout, stderr = process.communicate()
        if process.returncode != 0:
            logging.error('%s', stderr.decode())
            exit(-1)

def _combine_citations(lines):
    # convert \sphinxcite{A}\sphinxcite{B} to \sphinxcite{A,B}
    for i, l in enumerate(lines):
        if '}\sphinxcite{' in l:
            lines[i] = l.replace('}\sphinxcite{', ',')

# E.g., tag = 'begin{figure}'
def _tag_in_line(tag, line):
    assert '\\' not in set(tag)
    return any([elem.startswith(tag) for elem in line.split('\\')])

def _center_graphics(lines):
    tabulary_cnt = 0
    figure_cnt = 0
    in_doc = False
    for i, l in enumerate(lines):
        if _tag_in_line('begin{tabulary}', l):
            tabulary_cnt += 1
        elif _tag_in_line('end{tabulary}', l):
            tabulary_cnt -= 1
        elif _tag_in_line('begin{figure}', l):
            figure_cnt += 1
        elif _tag_in_line('end{figure}', l):
            figure_cnt -= 1
        elif _tag_in_line('begin{document}', l):
            in_doc = True

        # 'tabulary' and 'figure' blocks do include '\centering', so only center
        # '\sphinxincludegraphics{}' by enclosing it with '\begin{center}'
        # '\end{center}'. Logo should not be centered and it is not in_doc.
        if tabulary_cnt == 0 and figure_cnt == 0 and in_doc:
            sigs_greedy = re.findall('\\\\sphinxincludegraphics\\{.*\\}', l)
            if len(sigs_greedy) > 0:
                longest_balanced_braces = regex.findall('\{(?>[^{}]|(?R))*\}',
                                                        sigs_greedy[0])
                sig_with_balanced_braces = ('\\sphinxincludegraphics'
                                            + longest_balanced_braces[0])
                lines[i] = l.replace(sig_with_balanced_braces,
                                     ('\\begin{center}'
                                      + sig_with_balanced_braces
                                      + '\\end{center}'))



import re
from typing import Optional, List, Any, Callable, Tuple

# Our special mark in markdown, e.g. :label:`chapter_intro`
md_mark_pattern = re.compile(':([-\/\\._\w\d]+):(`[\ \*-\/\\\._\w\d]+`)?')
# Same for md_mark_pattern, but for rst files
rst_mark_pattern = re.compile(':([-\/\\._\w\d]+):(``[\ \*-\/\\\._\w\d]+``)?')
# The source code tab mark
source_tab_pattern = re.compile('# *@tab +([\w\,\ ]+)')

# Markdown code fence
md_code_fence = re.compile('(```+) *(.*)')

def group_list(list_obj: List[Any], status_fn: Callable[[Any, Any], Any]
              ) -> List[Tuple[Any, List[Any]]]:
    """Cut a list into multiple parts when fn returns True"""
    prev_status = None
    prev_pos = 0
    ret = []
    for i, item in enumerate(list_obj):
        cur_status = status_fn(item, prev_status)
        if prev_status is not None and cur_status != prev_status:
            ret.append((prev_status, list_obj[prev_pos:i]))
            prev_pos = i
        prev_status = cur_status
    ret.append((cur_status, list_obj[prev_pos:]))
    return ret



import configparser
import os
import logging

class Config():
    def __init__(self, tab=None, config_fname='config.ini'):
        config = configparser.ConfigParser()
        default_config_name = os.path.join(
            os.path.dirname(__file__), 'config_default.ini')
        config.read(default_config_name)
        if os.path.exists(config_fname):
            logging.info('Load configure from %s', config_fname)
            config.read(config_fname)
        tabs = config['build']['tabs']
        self.tabs = [tab.strip() for tab in tabs.lower().split(',')] if tabs else []
        self.default_tab = self.tabs[0] if self.tabs else None
        self.tab = tab.lower() if tab else self.default_tab
        if self.tab:
            assert self.tabs, 'No tabs is specified'
            if self.tab != 'all':
                assert self.tab in tabs, self.tab + ' is not found in tabs, which are ' + tabs
        self.build = config['build']
        self.deploy = config['deploy']
        self.project = config['project']
        self.html = config['html']
        self.pdf = config['pdf']
        self.library = config['library']
        self.colab = config['colab']
        self.sagemaker = config['sagemaker']

        # A bunch of directories
        self.src_dir = self.build['source_dir']
        self.tgt_dir = self.build['output_dir']
        self.eval_dir = os.path.join(self.tgt_dir, 'eval')
        self.ipynb_dir = os.path.join(self.tgt_dir, 'ipynb')
        self.rst_dir = os.path.join(self.tgt_dir, 'rst')
        try:
            self.html_dir = self.build['html_dir']
        except KeyError:
            self.html_dir = os.path.join(self.tgt_dir, 'html')
        # MM20200104 changed to allow separate html_dir to be specified in config.ini, e.g. put 'html_dir = docs' in the [build] section
        self.pdf_dir = os.path.join(self.tgt_dir, 'pdf')
        self.colab_dir = os.path.join(self.tgt_dir, 'colab')
        self.sagemaker_dir = os.path.join(self.tgt_dir, 'sagemaker')
        self.linkcheck_dir = os.path.join(self.tgt_dir, 'linkcheck')

        # Some targets names.
        self.pdf_fname = os.path.join(self.pdf_dir, self.project['name']+'.pdf')
        self.tex_fname = os.path.join(self.pdf_dir, self.project['name']+'.tex')
        self.pkg_fname = os.path.join(self.tgt_dir, self.project['name']+'.zip')

        # The project must have an index page
        index_fname, ext = os.path.splitext(self.build['index'])
        if ext and ext != '.md':
            logging.info('Ignore the file extesion, %s, specified by index in %s',
                         ext, config_fname)
        index_fname = os.path.join(self.src_dir, index_fname+'.md')
        if not os.path.exists(index_fname):
            logging.fatal('Failed to find the index file: %s', index_fname)
            exit(-1)

        if not self.project['title']:
            # Infer the book title from the index page
            with open(index_fname, 'r') as f:
                for line in f:
                    line = line.strip()
                    if line:
                        if line.startswith('#'):
                            line = line[1:]
                        self.project['title'] = line.strip()
                        break

        self.set_tab(self.tab)
        # Sanity checks.
        self.sanity_check()

    def sanity_check(self):
        notebook_patterns = self.build['notebooks'].split()
        for p in notebook_patterns:
            assert p.endswith('md'), '`notebooks` patterns must end with `md`' \
                   ' in `config.init`. Examples: `notebooks = *.md */*.md`.'

        rst_patterns = self.build['rsts'].split()
        for p in rst_patterns:
            assert p.endswith('rst'), '`rsts` patterns must end with `rst`' \
                    ' in `config.init`. Examples: `rsts = index.rst' \
                    ' api/**/*.rst`.'


    def iter_tab(self, fn):
        if self.tab == 'all':
            for tab in self.tabs:
                self.set_tab(tab)
                fn()
            self.set_tab('all')
        else:
            fn()

    def set_tab(self, tab):
        assert tab in self.tabs + ['all'], f"{tab} doesn't exist in {self.tabs}"
        self.tab = tab
        self.eval_dir = self._set_tab_dir(self.eval_dir, tab)
        self.ipynb_dir = self._set_tab_dir(self.ipynb_dir, tab)
        self.rst_dir = self._set_tab_dir(self.rst_dir, tab)
        self.pdf_dir = self._set_tab_dir(self.pdf_dir, tab)
        self.colab_dir = self._set_tab_dir(self.colab_dir, tab)
        self.sagemaker_dir = self._set_tab_dir(self.sagemaker_dir, tab)

    def _default_tab_dir(self, dirname):
        tokens = dirname.split('/')
        if self.tabs and '_' in tokens[-1]:
            tokens[-1] =  '_'.join(tokens[-1].split('_')[:-1])
            return '/'.join(tokens)
        return dirname

    def _set_tab_dir(self, dirname, tab):
        dirname = self._default_tab_dir(dirname)
        if tab == self.default_tab:
            return dirname
        return dirname + '_' + tab



import os
import logging
from d2lbook import sphinx_template as template
from d2lbook import utils

__all__ = ['prepare_sphinx_env']

def prepare_sphinx_env(config):
    env = SphinxEnv(config)
    env.prepare_env()

class SphinxEnv(object):
    def __init__(self, config):
        self.config = config
        self.pyconf = template.sphinx_conf

    def prepare_env(self):
        self._copy_static_files()
        self._update_header_links()
        self._write_js()
        self._write_css()
        for key in self.config.project:
            self._update_pyconf(key, self.config.project[key])
        self._update_pyconf('index', self.config.build['index'])
        self._update_pyconf('sphinx_configs', self.config.build['sphinx_configs'])

        extensions = ['recommonmark', 'sphinxcontrib.bibtex',
                      'sphinxcontrib.rsvgconverter', 'sphinx.ext.autodoc',
                      'sphinx.ext.viewcode']
        extensions.extend(self.config.build['sphinx_extensions'].split())
        self._update_pyconf('extensions', ','.join('"'+ext+'"' for ext in extensions))
        for font in ['main_font', 'sans_font', 'mono_font']:
            font_value = ''
            if self.config.pdf[font]:
                font_value = '\set%s{%s}' % (font.replace('_', ''), self.config.pdf[font])
            self._update_pyconf(font, font_value)

        fname = os.path.join(self.config.rst_dir, 'conf.py')
        with open(fname, 'w') as f:
            f.write(self.pyconf)

    def _update_pyconf(self, key, value):
        self.pyconf = self.pyconf.replace(key.upper(), value)

    def _copy_static_files(self):
        static_keys = [('html', 'favicon'), ('html', 'html_logo'), ('pdf', 'latex_logo')]
        for attribute, key in static_keys:
            if attribute == 'html':
                fname = self.config.html[key]
            elif attribute == 'pdf':
                fname = self.config.pdf[key]
            if not fname:
                self._update_pyconf(key, '')
                continue
            sphinx_fname = os.path.join(self.config.rst_dir, '_static',
                                        os.path.basename(fname))
            utils.copy(fname, sphinx_fname)
            self._update_pyconf(key, os.path.join(
                '_static', os.path.basename(fname)))

    def _update_header_links(self):
        items = utils.split_config_str(self.config.html['header_links'], 3)
        sphinx_links = ''
        for tk in items:
            if tk:
                sphinx_links += "('%s', '%s', True, '%s')," % (tk[0], tk[1], tk[2])
        self._update_pyconf('header_links', sphinx_links)

    def _write_js(self):
        d2l_js = (template.shorten_sec_num + template.replace_qr
                  + template.copybutton_js + template.discourse_js + template.tabbar_js)
        g_id = 'google_analytics_tracking_id'
        if g_id in self.config.deploy:
            d2l_js += template.google_tracker.replace(
                g_id.upper(), self.config.deploy[g_id])

        os.makedirs(os.path.join(self.config.rst_dir, '_static'), exist_ok=True)
        fname = os.path.join(self.config.rst_dir, '_static', 'd2l.js')
        with open(fname, 'w') as f:
            f.write(d2l_js)
            for fname in utils.find_files(self.config.html['include_js'], self.config.src_dir):
                with open (fname, 'r') as fin:
                    f.write(fin.read())

    def _write_css(self):
        fname = os.path.join(self.config.rst_dir, '_static', 'd2l.css')
        d2l_css = template.hide_bibkey_css + template.copybutton_css + \
            template.limit_output_length_css + template.tabbar_css
        with open(fname, 'w') as f:
            f.write(d2l_css)
            for fname in utils.find_files(self.config.html['include_css'], self.config.src_dir):
                with open (fname, 'r') as fin:
                    f.write(fin.read())



import os
import glob
import shutil
import logging

def rm_ext(filename):
    return os.path.splitext(filename)[0]

def find_files(pattern, root=None, excluded_pattern=None):
    fnames = []
    patterns = pattern.split()
    for p in patterns:
        if root is not None:
            p = os.path.join(root, p)
        if os.path.isdir(p):
            p = os.path.join(p, '**')
        for fn in glob.glob(p, recursive=True):
            if os.path.isfile(fn):
                fnames.append(fn)
    if not excluded_pattern:
        return fnames
    excluded_fnames = find_files(excluded_pattern, root)
    return [fn for fn in fnames if fn not in excluded_fnames]

def get_mtimes(fnames):
    if isinstance(fnames, str):
        return os.path.getmtime(fnames)
    return [os.path.getmtime(fn) for fn in fnames]

def split_fname(fname, base_dir, ext=None):
    fname = os.path.relpath(fname, base_dir)
    base, fext = os.path.splitext(fname)
    if fext.startswith('.'):
        fext = fext[1:]
    if ext and ext != fext:
        logging.warn("%s doesn't have extension %s", fname, ext)
    return base, fext

def get_tgt_fname(src_dir, src_fname, tgt_dir, src_ext, tgt_ext):
    fname, ext = split_fname(src_fname, src_dir, src_ext)
    if tgt_ext:
        ext = tgt_ext
    return os.path.join(tgt_dir, fname+'.'+ext)

def get_updated_files(src_fnames, src_dir, tgt_dir,
                      src_ext=None, tgt_ext=None, deps_mtime=0):
    updated_fnames = []
    for src_fn in src_fnames:
        tgt_fn = get_tgt_fname(src_dir, src_fn, tgt_dir, src_ext, tgt_ext)
        if (not os.path.exists(tgt_fn) # new
            or get_mtimes(src_fn) > get_mtimes(tgt_fn) # target is old
            or get_mtimes(tgt_fn) < deps_mtime): # deps is updated
            updated_fnames.append((src_fn, tgt_fn))
    return updated_fnames


def get_tgt_files_from_src_pattern(pattern, tgt_dir, src_ext, tgt_ext):
    """Get files with tgt_ext in tgt_dir according to pattern with src_ext"""
    patterns = pattern.split()
    for i, p in enumerate(patterns):
        f, ext = os.path.splitext(p)
        if src_ext and ext == '.' + src_ext and tgt_ext:
            patterns[i] = f + '.' + tgt_ext
    return find_files(' '.join(patterns), tgt_dir)


def get_files_to_rm(pattern, src_dir, tgt_dir, src_ext=None, tgt_ext=None):
    """Return files under tgt_dir whose corresponding src file is removed under src_dir."""
    tgt_files = get_tgt_files_from_src_pattern(pattern, tgt_dir, src_ext, tgt_ext)
    to_removes = []
    for tgt_fn in tgt_files:
        # If tgt_ext is provided, only files with tgt_ext in tgt_dir are
        # considered being removed. Note that ipynb to rst may generate svg
        # files, which should not be removed though these svg files do not have
        # corresponding files in src_dir
        if tgt_ext:
            fext = os.path.splitext(tgt_fn)[1]
            if fext.startswith('.'):
                fext = fext[1:]
            if tgt_ext != fext:
                continue
        # By switching args, it actually get_src_fname.
        src_fn = get_tgt_fname(tgt_dir, tgt_fn, src_dir, tgt_ext, src_ext)
        if not os.path.exists(src_fn):
            to_removes.append(tgt_fn)
    return to_removes


def rm_empty_dir(path, rmed_empty_dirs):
    """Recursively remove empty directories under and including path."""
    if not os.path.isdir(path):
        return

    fnames = os.listdir(path)
    if len(fnames) > 0:
        for fn in fnames:
            fpath = os.path.join(path, fn)
            if os.path.isdir(fpath):
                rm_empty_dir(fpath, rmed_empty_dirs)

    if len(os.listdir(path)) == 0:
        rmed_empty_dirs.append(str(path))
        os.rmdir(path)

def hide_individual_data_files(fns):
    """To display concisely: _build/eval/data/A/B/C/D -> _build/eval/data/A."""
    concise_fns = set()
    for fn in fns:
        concise_fn = []
        fn_components = fn.split('/')
        i = 0
        seen_data = False
        while i < len(fn_components) and not seen_data:
            component = fn_components[i]
            concise_fn.append(component)
            if component == 'data':
                seen_data = True
            i += 1
        if i < len(fn_components) - 1:
            next_component = fn_components[i + 1]
            if next_component.isdigit():
                concise_fn.append('<some digit>')
            else:
                concise_fn.append(next_component)
            if i < len(fn_components) - 2:
                concise_fn.append('')  # For indicating dir instead of file
        concise_fns.add('/'.join(concise_fn))
    return concise_fns

def mkdir(dirname):
    os.makedirs(dirname, exist_ok=True)


def copy(src, tgt):
    mkdir(os.path.dirname(tgt))
    shutil.copy(src, tgt)


def get_time_diff(tik, tok):
    h, remainder = divmod((tok - tik).seconds, 3600)
    m, s = divmod(remainder, 60)
    return "%02d:%02d:%02d" % (h, m, s)

def run_cmd(cmd, verbose=False):
    if isinstance(cmd, str):
        cmd = [cmd]
    cmd = ' '.join(cmd)
    if verbose:
        logging.info('Run "%s"', cmd)
    ret = os.system(cmd)
    if ret != 0:
        exit(-1)

def split_config_str(config_str, num_items_per_line=None):
    items = []
    if not config_str:
        return items
    lines = config_str.split('\n')
    for i, line in enumerate(lines):
        items.append([tk.strip() for tk in line.split(',') if tk.strip()])
        if num_items_per_line and len(items[-1]) != num_items_per_line:
            logging.fatal("The items in %d-th line (%d) doesn't"
                          " match the required (%d)"%(i, len(items[-1]), num_items_per_line))
            logging.fatal("The raw string is:\n"+config_str)
    return items



"""Integration with Sagemaker"""
import nbformat
from d2lbook import utils
from d2lbook import colab
from d2lbook import notebook

class Sagemaker(colab.Colab):
    def __init__(self, config):
        self._valid = config.sagemaker and config.sagemaker['github_repo']
        self.config = config.sagemaker
        self._repo, self._libs = colab.parse_repo_lib(
            self.config['github_repo'], self.config['libs'], config.library["version"])
        kernel_str = self.config['kernel']
        if ',' not in kernel_str:
            self._kernel = {None: kernel_str}
        else:
            kernel = utils.split_config_str(kernel_str, 2)
            self._kernel = {k[0]:k[1] for k in kernel}

    def generate_notebooks(self, eval_dir, sagemaker_dir, tab):
        if not self._valid:
            return
        utils.run_cmd(['rm -rf', sagemaker_dir])
        utils.run_cmd(['cp -r', eval_dir, sagemaker_dir])
        notebooks = utils.find_files('**/*.ipynb', sagemaker_dir)
        for fn in notebooks:
            nb = notebook.read(fn)
            if not nb:
                continue
            colab.update_notebook_kernel(nb, self._kernel[tab])
            colab.insert_additional_installation(nb, self._libs[tab], self.config['libs_header'])
            with open(fn, 'w') as f:
                f.write(nbformat.writes(nb))



"""utilities to handle rst files"""
import re
import logging
from typing import Dict
import nbconvert
import nbformat
from nbformat import notebooknode
from d2lbook import notebook
from d2lbook import common
from d2lbook import markdown

def convert_notebook(nb: notebooknode.NotebookNode, resources: Dict[str, str]):
    nb = _process_nb(nb)
    writer = nbconvert.RSTExporter()
    body, resources = writer.from_notebook_node(nb, resources)
    body = _process_rst(body)
    return body, resources

def _process_nb(nb):
    # add empty lines before and after a mark/fence
    new_cells = []
    for cell in nb.cells:
        if cell.cell_type == 'markdown':
            md_cells = markdown.split_markdown(cell.source)
            for i, md_cell in enumerate(md_cells):
                if i < len(md_cells) - 1 and md_cells[i+1]['type'] == 'code':
                    md_cells[i]['source'] += '\n'
                if md_cell['type'] == 'markdown':
                    lines = md_cells[i]['source'].split('\n')
                    for j, line in enumerate(lines):
                        m = common.md_mark_pattern.match(line)
                        if (m is not None
                            and m[1] not in ('ref', 'numref', 'eqref')
                            and m.end() == len(line)):
                            lines[j] = '\n'+line+'\n'
                    md_cells[i]['source'] = '\n'.join(lines)
            new_cells.append(nbformat.v4.new_markdown_cell(
                markdown.join_markdown_cells(md_cells)))
        else:
            new_cells.append(cell)
    # hide/show
    for cell in new_cells:
        if cell.cell_type == 'code':
            if '# hide outputs' in cell.source.lower():
                cell.outputs = []
            if '# hide code' in cell.source.lower():
                cell.source = ''
    return notebook.create_new_notebook(nb, new_cells)

def _process_rst(body):

    def delete_lines(lines, deletes):
        return [line for i, line in enumerate(lines) if i not in deletes]
    def indented(line):
        return line.startswith('   ')

    def blank(line):
        return len(line.strip()) == 0

    def look_behind(i, cond, lines):
        indices = []
        while i < len(lines) and cond(lines[i]):
            indices.append(i)
            i = i + 1
        return indices

    lines = body.split('\n')
    # deletes: indices of lines to be deleted
    i, deletes = 0, []
    while i < len(lines):
        line = lines[i]
        # '.. code:: toc' -> '.. toctree::', then remove consecutive empty lines
        # after the current line
        if line.startswith('.. code:: toc'):
            # convert into rst's toc block
            lines[i] = '.. toctree::'
            blanks = look_behind(i+1, blank, lines)
            deletes.extend(blanks)
            i += len(blanks)
        # .. code:: eval_rst
        #
        #
        #    .. only:: html
        #
        #       References
        #       ==========
        # ->
        #
        #
        #
        # .. only:: html
        #
        #    References
        #    ==========
        elif line.startswith('.. code:: eval_rst'):
            # make it a rst block
            deletes.append(i)
            j = i + 1
            while j < len(lines):
                line_j = lines[j]
                if indented(line_j):
                    lines[j] = line_j[3:]
                    if lines[j].strip().startswith('.. '):
                        lines[j] = '\n'+lines[j].strip()
                elif not blank(line_j):
                    break
                j += 1
            i = j
        elif line.startswith('.. parsed-literal::'):
            # add a output class so we can add customized css
            lines[i] += '\n    :class: output'
            i += 1
        # .. figure:: ../img/jupyter.png
        #    :alt: Output after running Jupyter Notebook. The last row is the URL
        #    for port 8888.
        #
        #    Output after running Jupyter Notebook. The last row is the URL for
        #    port 8888.
        #
        # :width:``700px``
        #
        # :label:``fig_jupyter``
        #->
        # .. _fig_jupyter:
        #
        # .. figure:: ../img/jupyter.png
        #    :width: 700px
        #
        #    Output after running Jupyter Notebook. The last row is the URL for
        #    port 8888.
        elif indented(line) and ':alt:' in line:
            # Image caption, remove :alt: block, it cause trouble for long captions
            caps = look_behind(i, lambda l: indented(l) and not blank(l), lines)
            deletes.extend(caps)
            i += len(caps)
        # .. table:: Dataset versus computer memory and computational power
        #    +-...
        #    |
        #    +-...
        #
        # :label:``tab_intro_decade``
        # ->
        # .. _tab_intro_decade:
        #
        # .. table:: Dataset versus computer memory and computational power
        #
        #    +-...
        #    |
        #    +-...
        #
        elif line.startswith('.. table::'):
            # Add indent to table caption for long captions
            caps = look_behind(i+1, lambda l: not indented(l) and not blank(l),
                               lines)
            for j in caps:
                lines[j] = '   ' + lines[j]
            i += len(caps) + 1
        else:
            i += 1

    # change :label:my_label: into rst format
    lines = delete_lines(lines, deletes)
    deletes = []

    for i, line in enumerate(lines):
        pos, new_line = 0, ''
        while True:
            match = common.rst_mark_pattern.search(line, pos)
            if match is None or match[2] is None:
                new_line += line[pos:]
                break
            start, end = match.start(), match.end()
            # e.g., origin=':label:``fig_jupyter``', key='label', value='fig_jupyter'
            origin, key, value = match[0], match[1], match[2]
            assert value.startswith('``') and value.endswith('``'), value
            value = value[2:-2]
            new_line += line[pos:start]
            pos = end
            # assert key in ['label', 'eqlabel', 'ref', 'numref', 'eqref', 'width', 'height'], 'unknown key: ' + key
            if key == 'label':
                new_line += '.. _' + value + ':'
            elif key in ['ref', 'numref', 'cite']:
                new_line += ':'+key+':`'+value+'`'
            elif key == 'eqref':
                new_line += ':eq:`'+value+'`'
            elif key in ['class', 'func']:
                new_line += ':py:'+key+':`'+value+'`'
            # .. math:: f
            #
            # :eqlabel:``gd-taylor``
            # ->
            # .. math:: f
            #    :label: gd-taylor
            elif key == 'eqlabel':
                new_line += '   :label: '+value
                if blank(lines[i-1]):
                    deletes.append(i-1)
            elif key in ['width', 'height']:
                new_line += '   :'+key+': '+value
            elif key == 'bibliography':
                # a hard coded plain bibtex style...
                new_line += ('.. bibliography:: ' + value +
                             '\n   :style: apa')
                             # '\n   :style: apa\n   :all:') MM 20200104 removed ':all:' so only the cited references get printed
            else:
                logging.fatal(f'unknown key {key}')

        lines[i] = new_line
    lines = delete_lines(lines, deletes)

    def move(i, j): # move line i to line j
        lines.insert(j, lines[i])
        if i > j:
            del lines[i+1]
        else:
            del lines[i]

    # move :width: or :width: just below .. figure::
    for i, line in enumerate(lines):
        if line.startswith('.. figure::'):
            for j in range(i+1, len(lines)):
                line_j = lines[j]
                if not indented(line_j) and not blank(line_j):
                    break
                if line_j.startswith('   :width:') or line_j.startswith('   :height:'):
                    move(j, i+1)

    # move .. _label: before a image, a section, or a table
    lines.insert(0, '')
    i = 0
    while i < len(lines):
        line = lines[i]
        if line.startswith('.. _'):
            for j in range(i-1, -1, -1):
                line_j = lines[j]
                if (line_j.startswith('.. table:')
                    or line_j.startswith('.. figure:')):
                    move(i, j-1)
                    lines.insert(j-1, '')
                    i += 1  # Due to insertion of a blank line
                    break
                if (len(set(line_j)) == 1
                    and line_j[0] in ['=','~','_', '-']):
                    k = max(j-2, 0)
                    move(i, k)
                    lines.insert(k, '')
                    i += 1  # Due to insertion of a blank line
                    break
        i += 1

    # change .. image:: to .. figure:: to they will be center aligned
    for i, line in enumerate(lines):
        if '.. image::' in line:
            lines[i] = line.replace('.. image::', '.. figure::')

    # sometimes the code results contains vt100 codes, widely used for
    # coloring, while it is not supported by latex.
    for i, l in enumerate(lines):
        lines[i] = re.sub(r'\x1b\[[\d;]*K', '',
                          re.sub(r'\x1b\[[\d;]*m', '', l))

    return '\n'.join(lines)



from d2lbook import markdown
import unittest

_markdown_src = r'''
# Test

first para

:begin_tab:`python2`
python is good
:end_tab:

another para

:eqref:`sec_1`

:begin_tab:`python 3`
python3 is better

```python 3
print(3)
```

:end_tab:


````bash
```bash
$ ls
```
````
'''

class TestMarkdown(unittest.TestCase):

    def test_split(self):
        cells = markdown.split_markdown(_markdown_src)
        self.assertEqual(len(cells), 7)
        self.assertEqual(cells[0]['type'], 'markdown')
        self.assertEqual(cells[1]['type'], 'markdown')
        self.assertEqual(cells[1]['class'], '`python2`')
        self.assertEqual(cells[3]['class'], '`python 3`')
        self.assertEqual(cells[5]['class'], 'bash')

    def test_merge(self):
        cells = markdown.split_markdown(_markdown_src)
        src = markdown.join_markdown_cells(cells)
        self.assertEqual(_markdown_src, src)

if __name__ == '__main__':
    unittest.main()



from d2lbook import notebook
from d2lbook import build
import unittest
import nbconvert


_markdown_src = r'''
# Test

first para

:begin_tab:`python2`
python is good
:end_tab:

another para

This is :eqref:`sec_1`

:begin_tab:`python2`
```python2
1+2+3
```
:end_tab:

:begin_tab:`python3`
python3 is better

- here
- haha

:end_tab:

```{.input .python}
1+2+3
```

```{.input .python}
#@tab python2
1+2+3
```

```bash
````
$ ls
````
```
'''

_multi_tab_cell = r'''
# Test

```{.input .python}
#@tab python2, python3
1+2
```

The end
'''

class TestNotebook(unittest.TestCase):

    def test_split_markdown_cell(self):
        nb = notebook.read_markdown(_markdown_src)
        new_nb = notebook.split_markdown_cell(nb)
        cells = new_nb.cells
        self.assertEqual(len(cells), 8)
        self.assertEqual(cells[0].cell_type, 'markdown')
        self.assertEqual(cells[1].cell_type, 'markdown')
        self.assertEqual(cells[1].metadata['tab'], ['python2'])
        self.assertEqual(cells[2].cell_type, 'markdown')
        self.assertEqual('tab' in cells[2].metadata, False)
        self.assertEqual(cells[3].metadata['tab'], ['python2'])
        self.assertEqual(cells[4].metadata['tab'], ['python3'])
        self.assertEqual(cells[5].cell_type, 'code')
        self.assertEqual(cells[6].cell_type, 'code')

    def test_get_tab_notebook(self):
        nb = notebook.split_markdown_cell(notebook.read_markdown(_markdown_src))
        new_nb = notebook.get_tab_notebook(nb, tab='python2', default_tab='python3')
        cells = new_nb.cells
        self.assertEqual(cells[0].cell_type, 'markdown')
        self.assertEqual(cells[1].cell_type, 'markdown')
        self.assertEqual(cells[1].metadata['tab'], 'python2')
        self.assertEqual(cells[2].cell_type, 'markdown')
        self.assertEqual('tab' in cells[2].metadata, False)
        self.assertEqual(cells[3].metadata['tab'], 'python2')
        self.assertEqual(cells[4].cell_type, 'code')
        self.assertEqual(cells[4].metadata['tab'], 'python2')
        self.assertEqual(len(cells), 6)

        new_nb = notebook.get_tab_notebook(nb, tab='python3', default_tab='python3')
        cells = new_nb.cells
        self.assertEqual(cells[3].metadata['tab'], 'python3')
        self.assertEqual(len(cells), 5)

        nb = notebook.read_markdown(_multi_tab_cell)
        cells = notebook.get_tab_notebook(nb, tab='python2', default_tab='python3').cells
        self.assertEqual(len(cells), 3)
        self.assertEqual(cells[1].metadata['tab'], 'python2')

        cells = notebook.get_tab_notebook(nb, tab='python3', default_tab='python3').cells
        self.assertEqual(len(cells), 3)
        self.assertEqual(cells[1].metadata['tab'], 'python3')

    def test_merge_tab_notebooks(self):
        nb = notebook.split_markdown_cell(notebook.read_markdown(_markdown_src))
        nb2 = notebook.get_tab_notebook(nb, tab='python2', default_tab='python3')
        nb3 = notebook.get_tab_notebook(nb, tab='python3', default_tab='python3')
        new_nb = notebook.merge_tab_notebooks([nb2, nb3])
        self.assertEqual(len(nb.cells), len(new_nb.cells))
        for cell, new_cell in zip(nb.cells, new_nb.cells):
            if new_cell.source != cell.source:
                self.assertTrue(new_cell.source in cell.source)

    def test_add_html_tab(self):
        nb = notebook.split_markdown_cell(notebook.read_markdown(_markdown_src))
        nb2 = notebook.get_tab_notebook(nb, tab='python2', default_tab='python3')
        nb3 = notebook.get_tab_notebook(nb, tab='python3', default_tab='python3')
        nb4 = notebook.merge_tab_notebooks([nb2, nb3])
        new_nb = notebook.add_html_tab(nb4, default_tab='python3')

        writer = nbconvert.RSTExporter()
        body, _ = writer.from_notebook_node(new_nb)
        # TODO(mli) add some asserts
        # print(build.process_rst(body))



if __name__ == '__main__':
    unittest.main()



from d2lbook import notebook
from d2lbook import rst
import unittest
import nbconvert

_markdown_src = r'''
# Test
:label:`test`

first para

python is good

another para

This is :eqref:`sec_1`

```python2
1+2+3
```

python3 is better

- here
- haha


```{.input .python}
1+2+3
```

```{.input .python}
#@tab python2
1+2+3
```

```bash
````
aa
````
```

## Section 2
:label:`sec_2`

```eval_rst
.. only:: html

   Table of Contents
   -----------------
```

```toc
:numbered:
:maxdepth: 2

install
user/index
develop/index
```

![Estimating the length of a foot](../img/koebel.jpg)
:width:`400px`
'''


class TestRst(unittest.TestCase):

    # TODO(mli) add some asserts
    def test_convert_notebook(self):
        nb = notebook.read_markdown(_markdown_src)
        body, _ = rst.convert_notebook(nb, {})
        # print(body)
